#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Virtual broker.

Note: portfolio balance will only change once a position is closed.
    
"""

from autotrader.brokers.virtual import utils

class Broker():
    
    def __init__(self, broker_config):
        self.fee                = 0
        self.leverage           = 1
        self.commission         = 0
        self.spread             = 0
        self.margin_available   = 0
        self.portfolio_balance  = 0
        self.pending_positions  = {}
        self.open_positions     = {}
        self.closed_positions   = {}
        self.cancelled_orders   = {}
        self.total_trades       = 0
        self.profitable_trades  = 0
        self.peak_value         = 0
        self.low_value          = 0
        self.max_drawdown       = 0
        self.base_currency      = 'AUD'
        self.NAV                = 0
    
    
    def place_order(self, order_details):
        ''' 
            Place order with broker. Currently only supports market 
            order types. 
            
        '''
        order_type  = order_details["order_type"]
        pair        = order_details["instrument"]
        size        = order_details["size"]
        order_price = order_details["price"]
        time        = order_details["order_time"]
        take_price  = order_details["take_profit"]
        HCF         = order_details["HCF"]
        stop_type   = order_details["stop_type"]
        
        if 'stop_distance' not in order_details:
            stop_distance   = None
            stop_price      = order_details["stop_loss"]
        else:
            stop_distance   = order_details['stop_distance']
            stop_price      = None
        
        order_no = self.total_trades + 1
        
        # Add position to self.pending_positions
        new_position = {'order_ID'  : order_no,
                        'type'      : order_type,
                        'pair'      : pair,
                        'order_time': time,
                        'order_price': order_price,
                        'stop'      : stop_price,
                        'take'      : take_price,
                        'size'      : size,
                        'HCF'       : HCF,
                        'distance'  : stop_distance,
                        'stop_type' : stop_type
                        }
        self.pending_positions[order_no] = new_position
        
        # Update trade tally
        self.total_trades += 1
    
    
    def open_position(self, order_no, candle):
        ''' Opens position with broker. '''
        
        # Calculate margin requirements
        current_price   = candle.Open
        pip_value       = utils.get_pip_ratio(self.pending_positions[order_no]['pair'])
        size            = self.pending_positions[order_no]['size']
        HCF             = self.pending_positions[order_no]['HCF']
        position_value  = abs(size) * current_price * HCF
        margin_required = self.calculate_margin(position_value)
        
        if size > 0:
            spread_cost = 0.5*self.spread * pip_value
        else:
            spread_cost = -0.5*self.spread * pip_value
        
        if margin_required < self.margin_available:
            # Fill order
            new_position = self.pending_positions[order_no]
            new_position['time_filled']     = candle.name
            new_position['entry_price']     = candle.Open + spread_cost
            self.open_positions[order_no]   = new_position
            
        else:
            self.cancelled_orders[order_no] = self.pending_positions[order_no]
    
    
    def update_positions(self, candle):
        ''' 
            Updates orders and open positions based on current candle. 
        '''
        # Tally for positions opened this update
        opened_positions = 0
        
        # Update pending positions
        for order_no in self.pending_positions:
            if self.pending_positions[order_no]['order_time'] != candle.name:
                if self.pending_positions[order_no]['type'] == 'market':
                    self.open_position(order_no, candle)
                    opened_positions += 1
                elif self.pending_positions[order_no]['type'] == 'limit':
                    if self.open_positions[order_no]['size'] > 0:
                        if candle.Low < self.open_positions[order_no]['order_price']:
                            self.open_position(order_no, candle)
                            opened_positions += 1
                    else:
                        if candle.High > self.open_positions[order_no]['order_price']:
                            self.open_position(order_no, candle)
                            opened_positions += 1
        
        
        # Remove position from pending positions
        if opened_positions > 0:
            for order_no in self.open_positions.keys():
                self.pending_positions.pop(order_no, 0)
            
            for order_no in self.cancelled_orders.keys():
                self.pending_positions.pop(order_no, 0)
        
        
        # Update trailing stops
        # For other methods, move the stop update to an external function
        # Can this be moved into the loop below?
        for order_no in self.open_positions:
            if self.open_positions[order_no]['stop_type'] == 'trailing_stop':
                # Trailing stop loss is enabled, check if price has moved SL
                
                if self.open_positions[order_no]['distance'] is not None:
                    pip_value = utils.get_pip_ratio(self.open_positions[order_no]['pair'])
                    pip_distance = self.open_positions[order_no]['distance']
                    distance = pip_distance*pip_value
                    
                else:
                    distance = abs(self.open_positions[order_no]['entry_price'] - \
                                   self.open_positions[order_no]['stop'])
                

                if self.open_positions[order_no]['size'] > 0:
                    # long position, stop loss only moves up
                    new_stop = candle.High - distance
                    if new_stop > self.open_positions[order_no]['stop']:
                        self.open_positions[order_no]['stop'] = new_stop
                    
                else:
                    # short position, stop loss only moves down
                    new_stop = candle.Low + distance
                    if new_stop < self.open_positions[order_no]['stop']:
                        self.open_positions[order_no]['stop'] = new_stop

        
        # Update self.open_positions
        open_position_orders = list(self.open_positions.keys())
        unrealised_PL  = 0        # Un-leveraged value
        for order_no in open_position_orders:
            if self.open_positions[order_no]['size'] > 0:
                if candle.Low < self.open_positions[order_no]['stop']:
                    # Stop loss hit
                    self.close_position(order_no, 
                                        candle, 
                                        self.open_positions[order_no]['stop']
                                        )
                elif candle.High > self.open_positions[order_no]['take']:
                    # Take Profit hit
                    self.close_position(order_no, 
                                        candle, 
                                        self.open_positions[order_no]['take']
                                        )
                else:
                    # Position is still open, update value of holding
                    size        = self.open_positions[order_no]['size']
                    entry_price = self.open_positions[order_no]['entry_price']
                    price       = candle.Close
                    HCF         = self.open_positions[order_no]['HCF']
                    position_value = size*(price - entry_price)*HCF
                    unrealised_PL += position_value
            
            else:
                if candle.High > self.open_positions[order_no]['stop']:
                    self.close_position(order_no, 
                                        candle, 
                                        self.open_positions[order_no]['stop']
                                        )
                elif candle.Low < self.open_positions[order_no]['take']:
                    self.close_position(order_no, 
                                        candle, 
                                        self.open_positions[order_no]['take']
                                        )
                else:
                    size        = self.open_positions[order_no]['size']
                    entry_price = self.open_positions[order_no]['entry_price']
                    price       = candle.Close
                    HCF         = self.open_positions[order_no]['HCF']
                    position_value = size*(price - entry_price)*HCF
                    unrealised_PL += position_value
        
        # Update margin available
        self.update_margin(candle.Close)
        
        # Update open position value
        self.NAV = self.portfolio_balance + unrealised_PL
        
    
    def close_position(self, order_no, candle, exit_price):
        ''' Closes positions. '''
        # Remove position from self.open_positions and calculate profit
        order_time  = self.open_positions[order_no]['order_time']
        order_price = self.open_positions[order_no]['order_price']
        order_type  = self.open_positions[order_no]['type']
        pair        = self.open_positions[order_no]['pair']
        entry_time  = self.open_positions[order_no]['time_filled']
        entry_price = self.open_positions[order_no]['entry_price']
        size        = self.open_positions[order_no]['size']
        stop_price  = self.open_positions[order_no]['stop']
        take_price  = self.open_positions[order_no]['take']
        HCF         = self.open_positions[order_no]['HCF']
        
        exit_time   = candle.name
        exit_price  = exit_price
        
        # Update portfolio with profit/loss
        gross_PL    = size*(exit_price - entry_price)*HCF
        open_trade_value    = abs(size)*entry_price*HCF
        close_trade_value   = abs(size)*exit_price*HCF
        commission  = (self.commission/100) * (open_trade_value + close_trade_value)
        net_profit  = gross_PL - commission
        
        self.add_funds(net_profit)
        if net_profit > 0:
            self.profitable_trades += 1
        
        # Add trade to closed positions
        closed_position = {'order_ID'       : order_no,
                           'type'           : order_type,
                           'pair'           : pair,
                           'order_time'     : order_time,
                           'order_price'    : order_price,
                           'entry_time'     : entry_time, 
                           'entry_price'    : entry_price,
                           'stop_price'     : stop_price,
                           'take_price'     : take_price,
                           'exit_time'      : exit_time,
                           'exit_price'     : exit_price,
                           'size'           : size,
                           'profit'         : net_profit,
                           'balance'        : self.portfolio_balance
                           }
        
        # Add to closed positions dictionary
        self.closed_positions[order_no] = closed_position
        
        # Remove position from open positions
        self.open_positions.pop(order_no, 0)
        
        # Update maximum drawdown
        self.update_MDD()
    
    
    def get_positions(self, pair=None):
        ''' Returns the open positions in the account. '''
        # Note that trades which get closed on the same candle that they are 
        # opened on will not be returned. 
        if pair is not None:
            open_positions = self.open_positions
            current_positions = [v for v in open_positions.values() if pair in v.values()]
            
        else:
            current_positions = self.open_positions
            
        return current_positions
    
    
    def add_funds(self, amount):
        ''' Add's funds to brokerage account. '''
        
        if self.portfolio_balance == 0:
            # If this is the initial deposit, set peak and low values for MDD
            self.peak_value = amount
            self.low_value = amount
            
        self.portfolio_balance  += amount
        # self.margin_available   += amount
    
    
    def get_balance(self):
        ''' Returns balance of account. '''
        return self.portfolio_balance
    
    
    def calculate_margin(self, position_value):
        ''' Calculates margin required to take a position. '''
        margin = position_value / self.leverage
        
        return margin
    
    
    def update_margin(self, close_price):
        ''' Updates margin available in account. '''
        
        # If the method below is too slow, just assume margin used is constant
        # during position, so that I dont have to iterate open pos. every 
        # candle
        margin_used = 0
        # open_position_orders = list(self.open_positions.keys())
        for order_no in self.open_positions:
            size            = self.open_positions[order_no]['size']
            HCF             = self.open_positions[order_no]['HCF']
            # HCF should be updated for current time.
            position_value  = abs(size) * close_price * HCF
            margin_required = self.calculate_margin(position_value)
            margin_used     += margin_required
        
        self.margin_available = self.portfolio_balance - margin_used


    def get_price(self, pair, data, conversion_data, i):
        ''' Returns the price data dict. '''
        
        quote_currency = pair[-3:]
        
        ask = data.Close[i]
        bid = data.Close[i]
        conversion_data = conversion_data.Close[i]
        
        if quote_currency == self.base_currency:
            negativeHCF = 1
            positiveHCF = 1
        else:
            negativeHCF = 1/conversion_data
            positiveHCF = 1/conversion_data
        
        price = {"ask": ask,
                 "bid": bid,
                 "negativeHCF": negativeHCF,
                 "positiveHCF": positiveHCF
                 }
        
        return price
    
    
    def update_MDD(self):
        ''' Function to calculate maximum portfolio drawdown '''
        balance     = self.portfolio_balance
        peak_value  = self.peak_value
        low_value   = self.low_value
        
        if balance > peak_value:
            self.peak_value = balance
            self.low_value = balance
            
        elif balance < low_value:
            self.low_value = balance
        
        MDD = 100*(low_value - peak_value)/peak_value
        # also need to initiate peak and low value to be equal to portfolio balance
            # Done in self.add_funds
        # Need to include time in equation
            # I think this is done implicitly 
        # Also need to check that the newest MDD is not less than previous MDD
        if MDD < self.max_drawdown:
            self.max_drawdown = MDD
            
        # Also include logic to do nothing if nothing changes for speed only
        # Isn't accurate yet, since if portfolio only goes up, low will be at
          # initial balance, giving a false MDD
          # Should only update low value if it comes after --- wait
          # I dont think this is true, ignore for now...

    def update_NAV(self):
        # Iterate over open positions to calculate value
        
        
        self.NAV = 0
        
            


# -*- coding: utf-8 -*-
"""
Custom indicators

"""
import talib as ta
import math
import numpy as np
import pandas as pd


def supertrend(data, period = 10, ATR_multiplier = 3.0):
    '''  '''
    hl2             = (data.High.values + data.Low.values) / 2
    true_range      = ta.TRANGE(data.High.values, data.Low.values, data.Close.values)
    atr             = ta.SMA(true_range, timeperiod=period)
    source          = hl2
    
    up              = source - (ATR_multiplier*atr)
    up_list         = [up[0]]
    up_times        = [data.index[0]]
    N_up            = 0
    
    dn              = source + (ATR_multiplier*atr)
    dn_list         = [dn[0]]
    dn_times        = [data.index[0]]
    N_dn            = 0
    
    trend           = 1
    trend_list      = [trend]
    
    for i in range(1, len(data)):
        
        if trend == 1:
            if data.Close.values[i] > max(up[N_up:i]):
                up_list.append(max(up[N_up:i]))
                up_times.append(data.index[i])
                
                dn_list.append(np.nan)
                dn_times.append(data.index[i])
                
            else: 
                trend = -1
                N_dn = i
                dn_list.append(dn[i])
                dn_times.append(data.index[i])
                
                up_list.append(np.nan)
                up_times.append(data.index[i])
                
        else:
            if data.Close.values[i] < min(dn[N_dn:i]):
                dn_list.append(min(dn[N_dn:i]))
                dn_times.append(data.index[i])
                
                up_list.append(np.nan)
                up_times.append(data.index[i])
                
            else:
                trend = 1
                N_up = i
                up_list.append(up[i])
                up_times.append(data.index[i])
                
                dn_list.append(np.nan)
                dn_times.append(data.index[i])
        
        trend_list.append(trend)
    
    up_trend = pd.DataFrame({'uptrend': up_list}, index = up_times)
    dn_trend = pd.DataFrame({'downtrend': dn_list}, index = dn_times)
    
    
    return up_trend, dn_trend, trend_list


def stoch_rsi(data, K_period=3, D_period=3, RSI_length=14, Stochastic_length=14):

    rsi1 = ta.RSI(data, timeperiod = RSI_length)
    stoch = stochastic(rsi1, rsi1, rsi1, Stochastic_length)
    
    K = sma(stoch, K_period)
    D = sma(K, D_period)
    
    return K, D


def stochastic(high, low, close, period=14):
    
    K = np.zeros(len(high))
    
    for i in range(period, len(high)):
        low_val     = min(low[i-period+1:i+1])
        high_val    = max(high[i-period+1:i+1])
        
        K[i]        = 100 * (close[i] - low_val)/(high_val - low_val)
        
    return K
    

def sma(data, period=14):
    
    sma_list = []
    
    for i in range(len(data)):
        average = sum(data[i-period+1:i+1])/period
        sma_list.append(average)
    
    return sma_list

# def ema(data, period):
    
#     sma_list = []
    
#     for i in range(len(data)):
#         average = sum(data[i-period+1:i+1])/period
#         sma_list.append(average)
    
#     return sma_list


def crossover(list_1, list_2):
    ''' Returns a list of length len(list_1) with 1 when list_1 crosses above
    list_2 and -1 when list_1 crosses below list_2.
    
    '''

    difference = [list_1[i] - list_2[i] for i in range(len(list_1))]
    sign_list = np.where(np.sign(difference) < 0, -1,1)
    
    crossover_list = [0]
    
    for i in range(1, len(sign_list)):
        if sign_list[i] - sign_list[i-1] != 0:
            val = sign_list[i]
        else:
            val = 0
        
        crossover_list.append(val)

    return crossover_list



def cross_values(a, b, ab_crossover):
        cross_point_list = [0]
        last_cross_point = 0
        for i in range(1, len(ab_crossover)):
            if ab_crossover[i] != 0:
                i0 = 0
                m_a = a[i] - a[i-1]
                m_b = b[i] - b[i-1]
                ix = (b[i-1] - a[i-1])/(m_a-m_b) + i0
                
                cross_point = m_a*(ix - i0) + a[i-1]
                
                last_cross_point = cross_point
                
            else:
                cross_point = last_cross_point #0
            
            cross_point_list.append(cross_point)
            
        return cross_point_list


def candles_between_crosses(cross_list):
    '''
    Returns candles since last cross
    
    
    Behaviour:
    in:  [0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1]
    out: [0, 0, 1, 1, 2, 3, 4, 1, 2, 3, 4, 5, 6]
    '''
    
    count = 0
    count_list = [count, count]
    
    for i in range(1, len(cross_list)-1):

        if cross_list[i] == 0:
            count += 1
        else:
            count = 1
        
        count_list.append(count)
    
    return count_list


def bullish_engulfing(data, detection = None):
    ''' Bullish engulfing pattern detection. '''
    
    if detection == "SMA50":
        sma50       = sma(data.Close.values, 50)
        
        down_trend  = np.where(data.Close.values < sma50, True, False)
        up_trend    = np.where(data.Close.values > sma50, True, False)
        
    elif detection == "SMA50/200":
        sma50       = sma(data.Close.values, 50)
        sma200      = sma(data.Close.values, 200)
        
        down_trend  = np.where((data.Close.values < sma50) & 
                               (data.Close.values < sma200), 
                               True, False)
        up_trend    = np.where((data.Close.values > sma50) & 
                               (data.Close.values > sma200), 
                               True, False)
    else:
        down_trend  = np.full(len(data), True)
        up_trend    = np.full(len(data), True)
    
    body_len        = 14    # ema depth for bodyAvg
    shadow_pc       = 100.0   # size of shadows 
    doji_pc         = 5.0
    shadow_factor   = 2.0 # number of times shadow dominates candle body
    
    body_high       = np.maximum(data.Close.values, data.Open.values)
    body_low        = np.minimum(data.Close.values, data.Open.values)
    body            = body_high - body_low
    
    body_avg        = ta.EMA(body, body_len)
    short_body      = body < body_avg
    long_body       = body > body_avg
    up_shadow       = data.High.values - body_high
    down_shadow     = body_low - data.Low.values
    
    has_up_shadow   = up_shadow > (shadow_pc / 100 * body)
    has_dn_shadow   = down_shadow > (shadow_pc / 100 * body)
    white_body      = data.Open.values < data.Close.values
    black_body      = data.Open.values > data.Close.values
    candle_range    = data.High.values - data.Low.values
    
    inside_bar = [False]
    for i in range(1, len(data)):
        val  = (body_high[i-1] > body_high[i]) and (body_low[i-1] < body_low[i])
        inside_bar.append(val)
        
    body_mid        = body/2 + body_low
    shadow_equals   = (up_shadow == down_shadow) | \
                          (
                              ((abs(up_shadow - down_shadow) / down_shadow * 100) < shadow_pc) & \
                              ((abs(down_shadow - up_shadow) / up_shadow * 100) < shadow_pc)
                          )
    doji_body       = (candle_range > 0) & (body <= candle_range * doji_pc / 100)
    doji            = doji_body & shadow_equals
    
    
    engulfing_bullish = [False]
    for i in range(1, len(data)):
        condition = down_trend[i] & \
                    white_body[i] & \
                    long_body[i] & \
                    black_body[i-1] & \
                    short_body[i-1] & \
                    (data.Close.values[i] >= data.Open.values[i-1]) & \
                    (data.Open.values[i] <= data.Close.values[i-1]) & \
                    ((data.Close.values[i] > data.Open.values[i-1]) | (data.Open.values[i] < data.Close.values[i-1]))
        
        engulfing_bullish.append(condition)
        
    return engulfing_bullish


def bearish_engulfing(data, detection = None):
    ''' Bearish engulfing pattern detection. '''
    
    if detection == "SMA50":
        sma50       = sma(data.Close.values, 50)
        
        down_trend  = np.where(data.Close.values < sma50, True, False)
        up_trend    = np.where(data.Close.values > sma50, True, False)
        
    elif detection == "SMA50/200":
        sma50       = sma(data.Close.values, 50)
        sma200      = sma(data.Close.values, 200)
        
        down_trend  = np.where((data.Close.values < sma50) & 
                               (data.Close.values < sma200), 
                               True, False)
        up_trend    = np.where((data.Close.values > sma50) & 
                               (data.Close.values > sma200), 
                               True, False)
    else:
        down_trend  = np.full(len(data), True)
        up_trend    = np.full(len(data), True)
    
    body_len        = 14    # ema depth for bodyAvg
    shadow_pc       = 100.0   # size of shadows 
    doji_pc         = 5.0
    shadow_factor   = 2.0 # number of times shadow dominates candle body
    
    body_high       = np.maximum(data.Close.values, data.Open.values)
    body_low        = np.minimum(data.Close.values, data.Open.values)
    body            = body_high - body_low
    
    body_avg        = ta.EMA(body, body_len)
    short_body      = body < body_avg
    long_body       = body > body_avg
    up_shadow       = data.High.values - body_high
    down_shadow     = body_low - data.Low.values
    
    has_up_shadow   = up_shadow > (shadow_pc / 100 * body)
    has_dn_shadow   = down_shadow > (shadow_pc / 100 * body)
    white_body      = data.Open.values < data.Close.values
    black_body      = data.Open.values > data.Close.values
    candle_range    = data.High.values - data.Low.values
    
    inside_bar = [False]
    for i in range(1, len(data)):
        val  = (body_high[i-1] > body_high[i]) and (body_low[i-1] < body_low[i])
        inside_bar.append(val)
        
    body_mid        = body/2 + body_low
    shadow_equals   = (up_shadow == down_shadow) | \
                          (
                              ((abs(up_shadow - down_shadow) / down_shadow * 100) < shadow_pc) & \
                              ((abs(down_shadow - up_shadow) / up_shadow * 100) < shadow_pc)
                          )
    doji_body       = (candle_range > 0) & (body <= candle_range * doji_pc / 100)
    doji            = doji_body & shadow_equals
    
    
    engulfing_bearish = [False]
    for i in range(1, len(data)):
        condition = up_trend[i] & \
                    black_body[i] & \
                    long_body[i] & \
                    white_body[i-1] & \
                    short_body[i-1] & \
                    (data.Close.values[i] <= data.Open.values[i-1]) & \
                    (data.Open.values[i] >= data.Close.values[i-1]) & \
                    ((data.Close.values[i] < data.Open.values[i-1]) | (data.Open.values[i] > data.Close.values[i-1]))
        
        engulfing_bearish.append(condition)
        
    return engulfing_bearish


def heikin_ashi(data):
    ''' 
        Calculates the Heikin-Ashi candlesticks from Japanese candlestick 
        data. 
    '''
    ha_close    = (data.Open + data.High + data.Low + data.Close)/4
    ha_open     = data.Open
    ha_low      = data.Low
    ha_high     = data.High
    
    for i in range(1, len(data)):
        ha_open[i]  = (data.Open[i-1] + data.Close[i-1])/2
        ha_low[i]   = min(data.Low[i], ha_open[i], ha_close[i])
        ha_high[i]  = max(data.High[i], ha_open[i], ha_close[i])
    
    ha_data = pd.concat([ha_open, ha_high, ha_low, ha_close], axis=1)
    ha_data.columns = ['Open', 'High', 'Low', 'Close']
    
    return ha_data
 

def find_swings(data, use_body = False):
    '''
        Locates the recent swings in price and returns a rolling list of 
        swing prices.
        
    '''
    # use_body flag currently not implemented. Idea is to use body close/open
    # instead of high/low (wicks) of candle for swings.
    
    # oc2     = (data.Open.values + data.Close.values)/2
    hl2     = (data.High.values + data.Low.values)/2
    n       = 2
    ema     = ta.EMA(hl2, n)
    
    grad    = np.gradient(ema)
    
    
    zeros   = np.where(np.sign(grad[1:len(grad)]) != np.sign(grad[0:-1]), 1, 0)
    zeros   = np.insert(zeros, 0, 0)
    swings  = -zeros*np.sign(grad)
    
    swing_df = pd.DataFrame(data=swings, index=data.index, columns=['swing'])
    
    low_list    = []
    high_list   = []
    for i in range(len(data)):
        if swing_df.swing[i] == -1:
            low_list.append(min(data.Low.values[i-1:i+2]))
            high_list.append(0)
        elif swing_df.swing[i] == 1:
            high_list.append(max(data.High.values[i-1:i+2]))
            low_list.append(0)
        else:
            low_list.append(0)
            high_list.append(0)
    
    swings_list     = merge_signals(low_list, high_list)
    last_swing      = rolling_signal_list(swings_list)
    last_swing[0:n] = list(data.High.values[0:n])
    
    # Need to return both a last swing low and last swing high list
    last_low        = rolling_signal_list(low_list)
    last_low[0:n]   = list(data.Low.values[0:n])
    last_high       = rolling_signal_list(high_list)
    last_high[0:n]  = list(data.High.values[0:n])
    
    swings          = pd.DataFrame(data={'Highs': last_high, 
                                         'Lows' : last_low,
                                         'Last' : last_swing},
                                   index = data.index)
    
    return swings
    

def rolling_signal_list(signals):
        ''' 
            Returns a list which maintains the previous signal, until a new 
            signal is given.
            
            [0,1,0,0,0,-1,0,0,1,0,0] ->  [0,1,1,1,1,-1,-1,-1,1,1,1]
            
        '''
        
        rolling_signals = [0]
        last_signal     = rolling_signals[0]
        
        for i in range(1, len(signals)):
            if signals[i] != 0:
                last_signal = signals[i]
            
            rolling_signals.append(last_signal)
        
        return rolling_signals


def merge_signals(signal_1, signal_2):
     ''' 
         Returns a single signal list which has merged two signal lists. 
     '''
     
     merged_signal_list = signal_1
     
     for i in range(len(signal_1)):
         if signal_2[i] != 0:
             merged_signal_list[i] = signal_2[i]
     
     return merged_signal_list


def ha_candle_run(ha_data):
    '''
        Returns a list for the number of consecutive green and red 
        Heikin-Ashi candles.
        
    '''
    green_candle    = np.where(ha_data.Close - ha_data.Open > 0, 1, 0)
    red_candle      = np.where(ha_data.Close - ha_data.Open < 0, 1, 0)
    
    green_run   = []
    red_run     = []
    
    green_sum   = 0
    red_sum     = 0
    
    for i in range(len(ha_data)):
        if green_candle[i] == 1:
            green_sum += 1
        else:
            green_sum = 0
        
        if red_candle[i] == 1:
            red_sum += 1
        else:
            red_sum = 0
        
        green_run.append(green_sum)
        red_run.append(red_sum)
        
    return green_run, red_run


# def TDI(data):
#     rsiPeriod = input(11, minval = 1, title = "RSI Period")
#     bandLength = input(31, minval = 1, title = "Band Length")
#     lengthrsipl = input(1, minval = 0, title = "Fast MA on RSI")
#     lengthtradesl = input(9, minval = 1, title = "Slow MA on RSI")
    
#     src = close                                                             // Source of Calculations (Close of Bar)
#     r = rsi(src, rsiPeriod)                                                 // RSI of Close
#     ma = sma(r, bandLength)                                                 // Moving Average of RSI [current]
#     offs = (1.6185 * stdev(r, bandLength))                                  // Offset
#     up = ma + offs                                                          // Upper Bands
#     dn = ma - offs                                                          // Lower Bands
#     mid = (up + dn) / 2                                                     // Average of Upper and Lower Bands
#     fastMA = sma(r, lengthrsipl)                                            // Moving Average of RSI 2 bars back
#     slowMA = sma(r, lengthtradesl)                                          // Moving Average of RSI 7 bars back
    
#     hline(30)                                                               // Oversold
#     hline(50)                                                               // Midline
#     hline(70)                                                               // Overbought
    
#     upl = plot(up, "Upper Band", color = blue)                              // Upper Band
#     dnl = plot(dn, "Lower Band", color = blue)                              // Lower Band
#     midl = plot(mid, "Middle of Bands", color = orange, linewidth = 2)      // Middle of Bands
    
#     plot(slowMA, "Slow MA", color=green, linewidth=2)                       // Plot Slow MA
#     plot(fastMA, "Fast MA", color=red, linewidth=2)                         // Plot Fast MA
    
#     fill(upl, midl, red, transp=90)                                         // Fill Upper Half Red
#     fill(midl, dnl, green, transp=90)                                       // Fill Lower Half Green






# -*- coding: utf-8 -*-
"""

Oanda API Wrapper
=================

"""

import v20
from autotrader.brokers.oanda import utils
import datetime
import pandas as pd


class Oanda():
    def __init__(self, oanda_config):
        # Create v20 context
        API             = oanda_config["API"]
        ACCESS_TOKEN    = oanda_config["ACCESS_TOKEN"]
        port            = oanda_config["PORT"]
        self.ACCOUNT_ID = oanda_config["ACCOUNT_ID"]
        self.api        = v20.Context(hostname=API, 
                                      token=ACCESS_TOKEN, 
                                      port=port)
        
        STREAM_API      = "stream-fxpractice.oanda.com"
        self.stream     = v20.Context(hostname=STREAM_API, 
                                      token=ACCESS_TOKEN, 
                                      port=443)
        
        self.open_positions     = {}
        
    
    def get_price(self, pair):
        
        response = self.api.pricing.get(accountID = self.ACCOUNT_ID, 
                                   instruments = pair
                                   )
        ask = response.body["prices"][0].closeoutAsk
        bid = response.body["prices"][0].closeoutBid
        negativeHCF = response.body["prices"][0].quoteHomeConversionFactors.negativeUnits
        positiveHCF = response.body["prices"][0].quoteHomeConversionFactors.positiveUnits
    
        price = {"ask": ask,
                 "bid": bid,
                 "negativeHCF": negativeHCF,
                 "positiveHCF": positiveHCF
                 }
    
        return price
    
    
    def place_order(self, order_details):
        ''' Places a market order with a stop loss and take profit. '''
        
        # Extract order details
        pair            = order_details["instrument"]
        stop_price      = order_details["stop_loss"]
        take_price      = order_details["take_profit"]
        size            = order_details["size"]
        
        # Check stop and take levels for precision requirements
        stop_price, take_price = self.check_precision(pair,
                                                      stop_price, 
                                                      take_price
                                                      )
        
        # Place order
        response = self.api.order.market(accountID = self.ACCOUNT_ID,
                              instrument = pair,
                              units = size,
                              takeProfitOnFill = {"price": str(take_price)},
                              stopLossOnFill = {"price": str(stop_price)},
                              )
        
        # Check response
        output = self.check_response(response)
        
        return response 
        

    def get_data(self, pair, period, interval):
        # print("Getting data for {}".format(pair))
        response    = self.api.instrument.candles(pair,
                                             granularity = interval,
                                             count = period,
                                             dailyAlignment = 0
                                             )
        
        data        = utils.response_to_df(response)
        
        return data
    
    
    def get_balance(self):
        response = self.api.account.get(accountID=self.ACCOUNT_ID)
        
        return response.body["account"].balance
    
    
    def get_positions(self, pair = None):
        ''' Gets the current positions open on the account. '''
        
        if pair is not None:
            
            response = self.api.position.get(accountID = self.ACCOUNT_ID, 
                                             instrument = pair)
            no_positions = 1
        else:
            response = self.api.position.list_open(accountID = self.ACCOUNT_ID)
            no_positions = len(response.body['positions'])
        
        return response.body
    
    
    def get_summary(self):
        
        # response = self.api.account.get(accountID=self.ACCOUNT_ID)
        response = self.api.account.summary(accountID=self.ACCOUNT_ID)
        # print(response.body['account'])
        
        return response
    
    
    def close_position(self, pair):
        ''' Closes all open positions on an instrument '''
        # Check if the position is long or short
        position    = self.get_positions(pair)['position']
        long_units  = position.long.units
        short_units = position.short.units
        
        if long_units > 0:
            response = self.api.position.close(accountID=self.ACCOUNT_ID, 
                                               instrument=pair, 
                                               longUnits="ALL")
            # # Check response
            # output = self.check_response(response)
        
        elif short_units > 0: 
            response = self.api.position.close(accountID=self.ACCOUNT_ID, 
                                               instrument=pair,
                                               shortUnits="ALL")
            # # Check response
            # output = self.check_response(response)
        
        else:
            print("There is no current position with {} to close.".format(pair))
            response = None
        
        return response
    
    
    def get_precision(self, pair):
        ''' Returns the allowable precision for a given pair '''
        response = self.api.account.instruments(accountID = self.ACCOUNT_ID, 
                                                instruments = pair)
        
        precision = response.body['instruments'][0].displayPrecision
        
        return precision
    
    
    def check_precision(self, pair, original_stop, original_take):
        ''' Modify stop/take based on pair for required ordering precision. ''' 
        N               = self.get_precision(pair)
        take_price      = round(original_take, N)
        stop_price      = round(original_stop, N)
        
        return stop_price, take_price
        
    
    def check_response(self, response):
        ''' Checks API response (currently only for placing orders) '''
        if response.status != 201:
            message = response.body['errorMessage']
        else:
            message = "Success."
            
        output = {'Status': response.status, 
                  'Message': message}
        
        return output
    
    def update_data(self, pair, granularity, data):
        ''' Attempts to construct the latest candle when there is a delay in the 
            api feed.
        '''
        
        granularity_details = self.deconstruct_granularity(granularity)
        secs = granularity_details['seconds']
        mins = granularity_details['minutes']
        hrs  = granularity_details['hours']
        days = granularity_details['days']
        
        # TODO: make this a function of original granularity
        # Done, but now I think, maybe this should always be something like S5?
        small_granularity = self.get_reduced_granularity(granularity_details, 
                                                         25)
        
        # Get data equivalent of last candle's granularity
        time_now        = datetime.datetime.now()
        start_time      = time_now - datetime.timedelta(seconds = secs,
                                                        minutes = mins,
                                                        hours = hrs,
                                                        days = days)
        latest_data     = self.get_historical_data(pair, 
                                                   small_granularity, 
                                                   start_time.timestamp(), 
                                                   time_now.timestamp())
        
        # Get latest price data
        latest_close    = latest_data.Close.values[0]
        
        open_price      = data.Close.values[-1]
        close_price     = latest_close
        high_price      = max(latest_data.High.values)
        low_price       = min(latest_data.Low.values)
        last_time       = data.index[-1]
        stripped_time   = datetime.datetime.strptime(last_time.strftime("%Y-%m-%d %H:%M:%S%z"),
                                                      "%Y-%m-%d %H:%M:%S%z")
        new_time        = stripped_time + datetime.timedelta(seconds = secs,
                                                              minutes = mins,
                                                              hours = hrs,
                                                              days = days)
        
        new_candle      = pd.DataFrame({'Open'  : open_price, 
                                        'High'  : high_price,
                                        'Low'   : low_price,
                                        'Close' : close_price},
                                        index=[new_time])
        
        
        new_data        = pd.concat([data, new_candle])
        
        return new_data
    
    
    def get_historical_data(self, pair, interval, from_time, to_time):
        response        = self.api.instrument.candles(pair,
                                                      granularity = interval,
                                                      fromTime = from_time,
                                                      toTime = to_time
                                                      )
        
        data = utils.response_to_df(response)
        
        return data
    
    
    def deconstruct_granularity(self, granularity):
        ''' Returns a dict with the seconds, minutes, hours and days
            corresponding to a granularity. 
        '''
        
        # Get letter to determine timeframe (eg. M)
        letter = granularity[0]
        
        # Get timeframe multiple (eg. 15)
        if len(granularity) > 1:
            number = float(granularity[1:])
        else:
            number = 1
        
        
        if letter == 'S':
            seconds     = number
            minutes     = 0
            hours       = 0
            days        = 0
        
        elif letter == 'M':
            seconds     = 0
            minutes     = number
            hours       = 0
            days        = 0
            
        elif letter == 'H':
            seconds     = 0
            minutes     = 0
            hours       = number
            days        = 0
            
        else:
            seconds     = 0
            minutes     = 0
            hours       = 0
            days        = number
        
        granularity_details = {'seconds': seconds,
                               'minutes': minutes,
                               'hours': hours,
                               'days': days}
        
        return granularity_details

    def get_reduced_granularity(self, granularity_details, fraction):
        '''Returns a candlestick granularity as a fraction of given granularity'''
        secs = granularity_details['seconds']
        mins = granularity_details['minutes']
        hrs  = granularity_details['hours']
        days = granularity_details['days']
        
        total_seconds = secs + 60*mins + 60*60*hrs + 60*60*24*days
        
        fractional_seconds = total_seconds/fraction
        
        seconds = fractional_seconds
        minutes = fractional_seconds/60
        hours = fractional_seconds/(60*60)
        days = fractional_seconds/(60*60*24)
        
        if days > 1:
            letter = 'D'
            number = 1
            reduced_granularity = letter
            
        elif hours > 1:
            base   = 2
            letter = 'H'
            number = base*round(hours/base)
            if number > 12:
                number = 12
            reduced_granularity = letter + str(number)
                
        elif minutes > 1:
            base   = 15 
            letter = 'M'
            number = base*round(minutes/base)
            if number > 30:
                number = 30
            reduced_granularity = letter + str(number)
            
        else: 
            base   = 15 
            letter = 'S'
            number = base*round(seconds/base)
            if number > 30:
                number = 30
            reduced_granularity = letter + str(number)
        
        if reduced_granularity[1:] == '0':
            reduced_granularity = reduced_granularity[0] + '1'
        
        return reduced_granularity


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
AutoTrader strategy template.
"""

# Import packages
import talib
import numpy as np
import plotly.graph_objs as go
from plotly.subplots import make_subplots
import plotly.io as pio
pio.renderers.default = 'browser'
import pandas as pd

# Path management
import os


class StrategyClass:

    def __init__(self, params, data, pair):
        ''' Define all indicators used in the strategy '''
        self.name   = "Strategy name"
        self.data   = data
        self.params = params
        
        # Construct indicators dict for plotting
        self.indicators = {}
        
        # Path variables
        strat_dir       = os.path.dirname(os.path.abspath(__file__))
        self.home_dir   = os.path.join(strat_dir, '..')
        

    def generate_signal(self, i, current_position):
        ''' Define strategy to determine entry signals '''
        
        # In this example, the strategy will place market orders.
        # Other order types are:
            # limit - to place a limit order
            # close - to close the current_position
        order_type  = 'market'
        signal_dict = {}
        
        # Put entry strategy here
        signal      = 0
        
        # Calculate exit targets
        exit_dict = self.generate_exit_levels(signal, i)
        
        # Construct signal dictionary
        signal_dict["order_type"]   = order_type
        signal_dict["direction"]    = signal
        signal_dict["stop_loss"]    = exit_dict["stop_loss"]
        signal_dict["stop_type"]    = exit_dict["stop_type"]
        signal_dict["take_profit"]  = exit_dict["take_profit"]
        
        return signal_dict
    
    
    def generate_exit_levels(self, signal, i):
        ''' Function to determine stop loss and take profit levels '''
        
        # Put exit strategy here
        stop = 0
        take = 0
        stop_type = 'limit'
        
        exit_dict = {'stop_loss'    : stop, 
                     'stop_type'    : stop_type,
                     'take_profit'  : take}
        
        return exit_dict
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Utility functions for Oanda.
----------------------------

"""

import pandas as pd
from datetime import datetime, timedelta
import numpy as np
import os

def response_to_df(response):
    ''' Function to convert api response into a pandas dataframe. '''
    
    candles = response.body["candles"]
    times = []
    close_price, high_price, low_price, open_price = [], [], [], []
    
    for candle in candles:
        if candle.complete == True:
            times.append(candle.time)
            close_price.append(float(candle.mid.c))
            high_price.append(float(candle.mid.h))
            low_price.append(float(candle.mid.l))
            open_price.append(float(candle.mid.o))
    
    dataframe = pd.DataFrame({"Open": open_price, "High": high_price, "Low": low_price, "Close": close_price})
    dataframe.index = pd.to_datetime(times)
    
    return dataframe


def truncate(f, n):
    ''' Truncates a float f to n decimal places without rounding. '''
    s = '{}'.format(f)
    
    if 'e' in s or 'E' in s:
        return '{0:.{1}f}'.format(f, n)
    i, p, d = s.partition('.')
    
    return '.'.join([i, (d+'0'*n)[:n]])


def get_pip_ratio(pair):
    ''' Function to return pip value ($/pip) of a given pair. '''
    if pair[-3:] == 'JPY':
        pip_value = 1e-2
    else:
        pip_value = 1e-4
    
    return pip_value


def get_size(pair, amount_risked, price, stop_price, HCF):
    ''' Calculate position size based on account balance and risk profile. '''
    pip_value           = get_pip_ratio(pair)
    pip_stop_distance   = abs(price - stop_price) / pip_value
    quote_risk          = amount_risked / HCF
    price_per_pip       = quote_risk / pip_stop_distance    
    units               = round(price_per_pip / pip_value)
    
    return units


def check_precision(pair, original_stop, original_take):
    ''' Modify stop/take based on pair for required ordering precision. ''' 
    if pair[-3:] == 'JPY':
        N = 3
    else:
        N = 5
    
    take_price      = float(truncate(original_take, N))
    stop_price      = float(truncate(original_stop, N))
    
    return stop_price, take_price

def update_data_with_candle(data, latest_candle):
    ''' Appends the latest candle to the data. '''
    # latest candle will be in the form of a list.
    # [time, Open, High, Low, Close]
    # Need to convert this to a df.
    # latest_candle = ['2021-06-14 02:54:30', 1.20994, 1.20994, 1.20994, 1.20994]
    candle_time = datetime.strptime(latest_candle[0],
                                    '%Y-%m-%d %H:%M:%S')
    
    candle = pd.DataFrame({"Open": latest_candle[1], 
                      "High": latest_candle[2], 
                      "Low": latest_candle[3],
                      "Close": latest_candle[4]},
                      index=[candle_time])
    new_data = pd.concat([data, candle])
    
    return new_data

def last_period(current_time, granularity):
    ''' 
        Returns a datetime object corresponding to the last candle that closed,
        based on the current time and the granularity of the candle.
        
        Note: if the current time is 10:15:01, the instrument will have just 
        begun forming the 10:15 candle. As such, the latest candle in the 
        instruments price history is the 10:00 candle, as that is the last 
        candle that has closed.
        
    '''    
    
    letter_to_unit = {'S': 'second',
                      'M': 'minute',
                      'H': 'hour',
                      'D': 'day'}
    
    letter = granularity[0]
    if len(granularity) > 1:
        number = int(granularity[1:])
    else:
        number = 1
    
    current_period      = getattr(current_time, letter_to_unit[letter])
    last_period         = number * np.floor(current_period/number) - number
    
    if letter == 'S':
        td = timedelta(microseconds = current_time.microsecond,
                       seconds = current_time.second - last_period)
    elif letter == 'M':
        td = timedelta(microseconds = current_time.microsecond,
                       seconds = current_time.second,
                       minutes = current_time.minute - last_period)
    elif letter == 'H':
        td = timedelta(microseconds = current_time.microsecond,
                       seconds = current_time.second,
                       minutes = current_time.minute,
                       hours = current_time.hour - last_period)
    elif letter == 'H':
        td = timedelta(microseconds = current_time.microsecond,
                       seconds = current_time.second,
                       minutes = current_time.minute,
                       hours = current_time.hour,
                       days = current_time.day - last_period)
    
    
    last_candle_closed = current_time - td
    
    
    return last_candle_closed

def write_to_order_summary(order_details, filepath):
    ''' Writes order details to summary file. '''
    
    # Check if file exists already, if not, create
    if not os.path.exists(filepath):
        f = open(filepath, "w")
        f.write("order time, strategy, granularity, order_type, instrument, order_size, ")
        f.write("trigger_price, stop_loss, take_profit\n")
        f.close()
    
    order_time          = order_details["order_time"]
    strategy            = order_details["strategy"]
    order_type          = order_details["order_type"]
    instrument          = order_details["instrument"]
    size                = order_details["size"]
    trigger_price       = order_details["price"]
    stop_loss           = order_details["stop_loss"]
    take_profit         = order_details["take_profit"]
    granularity         = order_details["granularity"]
    
    f                   = open(filepath, "a")
    f.write("{}, {}, {}, {}, {}, {}, {}, {}, {}\n".format(order_time, strategy, 
                                                          granularity, order_type, 
                                                          instrument, size, 
                                                          trigger_price, stop_loss, 
                                                          take_profit))
    f.close()
    
    
    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Utilities for virtual broker.

"""

import pandas as pd
import numpy as np
import talib as ta

def response_to_df(response):
    ''' Function to convert api response into a pandas dataframe. '''
    
    candles = response.body["candles"]
    times = []
    close_price, high_price, low_price, open_price = [], [], [], []
    
    for candle in candles:
        times.append(candle.time)
        close_price.append(float(candle.mid.c))
        high_price.append(float(candle.mid.h))
        low_price.append(float(candle.mid.l))
        open_price.append(float(candle.mid.o))
    
    dataframe = pd.DataFrame({"Open": open_price, "High": high_price, "Low": low_price, "Close": close_price})
    dataframe.index = pd.to_datetime(times)
    
    return dataframe


def truncate(f, n):
    ''' Truncates a float f to n decimal places without rounding. '''
    s = '{}'.format(f)
    
    if 'e' in s or 'E' in s:
        return '{0:.{1}f}'.format(f, n)
    i, p, d = s.partition('.')
    
    return '.'.join([i, (d+'0'*n)[:n]])


def get_pip_ratio(pair):
    ''' Function to return pip value ($/pip) of a given pair. '''
    if pair[-3:] == 'JPY':
        pip_value = 1e-2
    else:
        pip_value = 1e-4
    
    return pip_value


def get_size(pair, amount_risked, price, stop_price, HCF):
    ''' Calculate position size based on account balance and risk profile. '''
    if np.isnan(stop_price):
        units = 10000 # TODO: generalise
    else:
        pip_value           = get_pip_ratio(pair)
        pip_stop_distance   = abs(price - stop_price) / pip_value
        if pip_stop_distance == 0:
            units           = 0
        else:
            quote_risk      = amount_risked / HCF
            price_per_pip   = quote_risk / pip_stop_distance
            units           = round(price_per_pip / pip_value)
    
    return units


def check_precision(pair, original_stop, original_take):
    ''' Modify stop/take based on pair for required ordering precision. ''' 
    if pair[-3:] == 'JPY':
        N = 3
    else:
        N = 5
    
    take_price      = float(truncate(original_take, N))
    stop_price      = float(truncate(original_stop, N))
    
    return stop_price, take_price


def interval_to_seconds(interval):
    '''Converts the interval to time in seconds'''
    letter = interval[0]
    
    if len(interval) > 1:
        number = float(interval[1:])
    else:
        number = 1
    
    conversions = {'S': 1,
                   'M': 60,
                   'H': 60*60,
                   'D': 60*60*24
                   }
    
    my_int = conversions[letter] * number
    
    return my_int


def create_trade_lists(pair, timestamps, closed_positions_dict):
    
    long_entry_times    = []
    long_entry_prices   = []
    long_stops          = []
    long_takes          = []
    
    short_entry_times   = []
    short_entry_prices  = []
    short_stops         = []
    short_takes         = []
    
    portfolio_balance   = []
    portfolio_times     = []
    
    for order in closed_positions_dict:
        if closed_positions_dict[order]['pair'] == pair:
            portfolio_balance.append(closed_positions_dict[order]['balance'])      
            portfolio_times.append(closed_positions_dict[order]['exit_time'])
            
            if closed_positions_dict[order]['size'] > 0:
                long_entry_times.append(closed_positions_dict[order]['entry_time'])
                long_entry_prices.append(closed_positions_dict[order]['entry_price'])
                long_stops.append(closed_positions_dict[order]['stop_price'])
                long_takes.append(closed_positions_dict[order]['take_price'])
            else:
                short_entry_times.append(closed_positions_dict[order]['entry_time'])
                short_entry_prices.append(closed_positions_dict[order]['entry_price'])
                short_stops.append(closed_positions_dict[order]['stop_price'])
                short_takes.append(closed_positions_dict[order]['take_price'])

    # TODO all of these could just be put into a dataframe...
    return long_entry_times, long_entry_prices, \
           long_stops, long_takes, short_entry_times, \
           short_entry_prices, short_stops, short_takes, \
           portfolio_balance, portfolio_times



def fix_dataframe(df1, df2):
    # Would be good to check which one is shorter, which is longer, then 
    # return both with corrections
    
    i1 = list(df1.index)
    i2 = list(df2.index)
    new_indices = list(set(i1) - set(i2))
    
    new_df = df2
    
    for index in new_indices:
        
        df_row = df1[df1.index == index]
        
        df_row.Open = None
        df_row.High = None
        df_row.Low = None
        df_row.Close = None
        
        new_df = new_df.append(df_row)
    
    new_df = new_df.sort_index()
    new_df = new_df.interpolate()
    
    return new_df
    
    # insert the row with new index
    # Sort df in ascending order to move new idices into position
    # interpolate to get values at new rows



def check_dataframes(df_1, df_2):
    '''Checks dataframe lengths and corrects if necessary'''
    
    if len(df_1) < len(df_2):
        new_df_1 = fix_dataframe(df_2, df_1)
        new_df_2 = df_2
    elif len(df_1) > len(df_2):
        new_df_2 = fix_dataframe(df_1, df_2)
        new_df_1 = df_1
        
    else:
        new_df_1 = df_1
        new_df_2 = df_2
    
    return new_df_1, new_df_2


def trade_summary(pair, closed_positions_dict):
    # Could also include a pairs list, and only input closed_pos dict
    order_ID    = []
    times_list  = []
    order_price = []
    entry_time  = []
    entry_price = []
    size        = []
    stop_price  = []
    take_price  = []
    profit      = []
    portfolio_balance = []
    exit_times  = []
    exit_prices = []
    trade_duration = []
    
    for order in closed_positions_dict:
        if closed_positions_dict[order]['pair'] == pair:
            order_ID.append(closed_positions_dict[order]['order_ID'])
            entry_time.append(closed_positions_dict[order]['entry_time'])
            times_list.append(closed_positions_dict[order]['order_time'])
            order_price.append(closed_positions_dict[order]['order_price'])
            entry_price.append(closed_positions_dict[order]['entry_price'])
            size.append(closed_positions_dict[order]['size'])
            stop_price.append(closed_positions_dict[order]['stop_price'])
            take_price.append(closed_positions_dict[order]['take_price'])
            profit.append(closed_positions_dict[order]['profit'])
            portfolio_balance.append(closed_positions_dict[order]['balance'])
            exit_times.append(closed_positions_dict[order]['exit_time'])
            exit_prices.append(closed_positions_dict[order]['exit_price'])
            trade_duration.append(closed_positions_dict[order]['exit_time'].timestamp() - 
                                  closed_positions_dict[order]['entry_time'].timestamp())
            
    dataframe = pd.DataFrame({"Order_ID": order_ID, "Order_price": order_price,
                              "Entry_time": entry_time,
                              "Entry": entry_price, "Size": size,
                              "Stop_loss": stop_price, "Take_profit": take_price,
                              "Profit": profit, "Balance": portfolio_balance,
                              "Exit_time": exit_times, "Exit_price": exit_prices,
                              "Trade_duration": trade_duration})
    dataframe.index = pd.to_datetime(times_list)
    dataframe = dataframe.sort_index()
    
    return dataframe

def reconstruct_portfolio(initial_balance, trade_summary, time_index):
    a = trade_summary.Exit_time.values
    
    profit_list = trade_summary.Profit.values.tolist()
    new_df = pd.DataFrame({"Profit": profit_list})
    new_df.index = pd.to_datetime(a, utc=True)
    
    balance   = initial_balance
    portfolio = []
    
    for timestamp in time_index:
        if timestamp in new_df.index:
            # Check if timestamp appears in multiple closed positions
            profit_vals = new_df.Profit[new_df.index == timestamp].values
            profit = 0
            for trade_profit in profit_vals:
                profit += trade_profit
            
        else:
            profit = 0
        
        balance += profit
        portfolio.append(balance)
    
    dataframe = pd.DataFrame({"Balance": portfolio})
    dataframe.index = pd.to_datetime(time_index)
    dataframe = dataframe.sort_index()
    
    return dataframe


def trade_correlation(data, trade_summary):
    
    # Calculate indicators
    RSI         = ta.RSI(data.Close, timeperiod=14)
    ADX14       = ta.ADX(data.High, data.Low, data.Close, timeperiod=14)
    ADX200      = ta.ADX(data.High, data.Low, data.Close, timeperiod=200)
    sk, sd      = ta.STOCH(data.High, data.Low, data.Close)
    fk, fd      = ta.STOCHF(data.High, data.Low, data.Close)
    
    # Extract relevant data
    RSI_traded  = []
    ADX14_traded  = []
    ADX200_traded  = []
    sk_traded   = []
    sd_traded   = []
    fk_traded   = []
    fd_traded   = []
    
    for timestamp in data.index:
        if timestamp in trade_summary.index:
            RSI_traded.append(RSI[RSI.index == timestamp].values[0])
            ADX14_traded.append(ADX14[ADX14.index == timestamp].values[0])
            ADX200_traded.append(ADX200[ADX200.index == timestamp].values[0])
            sk_traded.append(sk[sk.index == timestamp].values[0])
            sd_traded.append(sd[fk.index == timestamp].values[0])
            fk_traded.append(fk[RSI.index == timestamp].values[0])
            fd_traded.append(fd[RSI.index == timestamp].values[0])
    
    # Construct correlation dataframe
    indicator_df = pd.DataFrame({"RSI": RSI_traded, "ADX_14": ADX14_traded,
                                 "ADX_200": ADX200_traded})
    indicator_df.index = trade_summary.index
    
    extended_trade_summary = trade_summary.merge(indicator_df,
                                                 left_index=True,
                                                 right_index=True)
    
    return extended_trade_summary

# def get_correlation(extended_trade_summary, indicator, ):
    
def get_streaks(trade_summary):
    
    profit_list             = trade_summary.Profit.values
    longest_winning_streak  = 1
    longest_losing_streak   = 1
    streak                  = 1
    
    for i in range(1, len(profit_list)):
        if np.sign(profit_list[i]) == np.sign(profit_list[i-1]):
            streak += 1
            
            if np.sign(profit_list[i]) > 0:
                # update winning streak
                longest_winning_streak  = max(longest_winning_streak, streak)
            else:
                # Update losing 
                longest_losing_streak   = max(longest_losing_streak, streak)

        else:
            streak = 1
    
    return longest_winning_streak, longest_losing_streak




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
''' 

Simple MACD strategy. 
------------------------
Rules for strategy:
    1. Trade in direction of trend, as per 200EMA.
    2. Entry signal on MACD cross below/above zero line.
    3. Set stop loss at recent price swing.
    4. Target 1.5 take profit.

'''

# Import packages
import talib
from autotrader.lib import indicators

# Path management
import os


class SimpleMACD:

    def __init__(self, params, data, pair):
        ''' Define all indicators used in the strategy '''
        self.name   = "Simple MACD Trend Strategy"
        self.data   = data
        self.params = params
        
        # 200EMA
        self.ema    = talib.EMA(data.Close.values, params['ema_period'])
        
        # MACD
        self.MACD, self.MACDsignal, self.MACDhist = talib.MACD(data['Close'].values, 
                                                  self.params['MACD_fast'], 
                                                  self.params['MACD_slow'], 
                                                  self.params['MACD_smoothing']
                                                  )
        self.MACD_CO        = indicators.crossover(self.MACD, self.MACDsignal)
        self.MACD_CO_vals   = indicators.cross_values(self.MACD, 
                                                      self.MACDsignal,
                                                      self.MACD_CO)
        # Construct indicators dict for plotting
        self.indicators = {'MACD (12/26/9)': {'type': 'MACD',
                                              'macd': self.MACD,
                                              'signal': self.MACDsignal,
                                              'histogram': self.MACDhist},
                           'EMA (200)': {'type': 'EMA',
                                         'data': self.ema},
                           'RSI (14)': {'type': 'RSI',
                                        'data': talib.RSI(data.Close.values)},
                           'RSI (7)': {'type': 'RSI',
                                       'data': talib.RSI(data.Close.values, 7)},
                           'EMA (21)': {'type': 'EMA',
                                        'data': talib.EMA(data.Close.values, 21)}}
        
        # Price swings
        self.swings         = indicators.find_swings(data)
        
        # Path variables
        strat_dir       = os.path.dirname(os.path.abspath(__file__))
        self.home_dir   = os.path.join(strat_dir, '..')
        

    def generate_signal(self, i, current_position):
        ''' Define strategy to determine entry signals '''
        
        order_type  = 'market'
        signal_dict = {}
        
        if self.data.Close.values[i] > self.ema[i] and \
            self.MACD_CO[i] == 1 and \
            self.MACD_CO_vals[i] < 0:
                signal = 1
                
        elif self.data.Close.values[i] < self.ema[i] and \
            self.MACD_CO[i] == -1 and \
            self.MACD_CO_vals[i] > 0:
                signal = -1

        else:
            signal = 0
        
        # Calculate exit targets
        exit_dict = self.generate_exit_levels(signal, i)
        
        # Construct signal dictionary
        signal_dict["order_type"]   = order_type
        signal_dict["direction"]    = signal
        signal_dict["stop_loss"]    = exit_dict["stop_loss"]
        signal_dict["stop_type"]    = exit_dict["stop_type"]
        signal_dict["take_profit"]  = exit_dict["take_profit"]
        
        return signal_dict
    
    
    def generate_exit_levels(self, signal, i):
        ''' Function to determine stop loss and take profit levels '''
        
        stop_type   = 'limit'
        RR          = self.params['RR']
        
        if signal == 0:
            stop    = None
            take    = None
        else:
            if signal == 1:
                stop    = self.swings.Lows[i]
                take    = self.data.Close[i] + RR*(self.data.Close[i] - stop)
            else:
                stop    = self.swings.Highs[i]
                take    = self.data.Close[i] - RR*(stop - self.data.Close[i])
                
        
        exit_dict   = {'stop_loss'    : stop, 
                       'stop_type'    : stop_type,
                       'take_profit'  : take}
        
        return exit_dict
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
AutoPlot
----------
General plotting script.

I want to be able to specify my indicators, time period and instrument, and 
get a plot for it.

"""

from datetime import timedelta
import pandas as pd
import plotly.graph_objs as go
from plotly.subplots import make_subplots
import plotly.io as pio
pio.renderers.default = 'browser'
from autotrader.lib import environment_manager
from autotrader.lib.autodata import GetData
import yaml
import numpy as np
from getopt import getopt
import sys
import os


# Bokeh
from bokeh.models.annotations import Title
from bokeh.plotting import figure, output_file, show, save
from bokeh.models import (
    CustomJS,
    ColumnDataSource,
    HoverTool
)
from bokeh.layouts import gridplot
from bokeh.transform import factor_cmap
import os


def read_yaml(file_path):
    '''Function to read and extract contents from .yaml file.'''
    with open(file_path, "r") as f:
        return yaml.safe_load(f)


def current_candle(current_time, granularity):
    ''' 
        Returns a datetime object corresponding to the last candle that closed,
        based on the current time and the granularity of the candle.
        
        Note: if the current time is 10:15:01, the instrument will have just 
        begun forming the 10:15 candle. As such, the latest candle in the 
        instruments price history is the 10:00 candle, as that is the last 
        candle that has closed.
        
    '''
    
    letter_to_unit = {'S': 'second',
                      'M': 'minute',
                      'H': 'hour',
                      'D': 'day'}
    
    letter = granularity[0]
    if len(granularity) > 1:
        number = int(granularity[1:])
    else:
        number = 1
    
    current_period      = getattr(current_time, letter_to_unit[letter])
    last_period         = number * np.floor(current_period/number)# - number
    
    if letter == 'S':
        td = timedelta(microseconds = current_time.microsecond,
                       seconds = current_time.second - last_period)
    elif letter == 'M':
        td = timedelta(microseconds = current_time.microsecond,
                       seconds = current_time.second,
                       minutes = current_time.minute - last_period)
    elif letter == 'H':
        td = timedelta(microseconds = current_time.microsecond,
                       seconds = current_time.second,
                       minutes = current_time.minute,
                       hours = current_time.hour - last_period)
    elif letter == 'H':
        td = timedelta(microseconds = current_time.microsecond,
                       seconds = current_time.second,
                       minutes = current_time.minute,
                       hours = current_time.hour,
                       days = current_time.day - last_period)
    
    
    last_candle_closed = current_time - td
    
    return last_candle_closed


def granularity_to_seconds(granularity):
    '''Converts the interval to time in seconds'''
    letter = granularity[0]
    
    if len(granularity) > 1:
        number = float(granularity[1:])
    else:
        number = 1
    
    conversions = {'S': 1,
                   'M': 60,
                   'H': 60*60,
                   'D': 60*60*24
                   }
    
    my_int = conversions[letter] * number
    
    return my_int


def plot_oanda_trade_history(filepath, global_config, broker_config, 
                             granularity, instruments=None):
    """
        Code below takes oanda csv history and plots it. Pretty cool!
        
        When a position is opened, it will be shown as a triangle with no fill (open).
        It will point in the direction of the trade.
        
        When a position is closed, it will be shown as a triangle with red or green 
        fill, depending on if it was profitable or not. It will point in the direction 
        of the closing trade.
        
    """
    import AutoTrader
    
    trade_history = pd.read_csv(filepath, index_col = 0)
    trade_history = trade_history.fillna(method='ffill')
    
    if instruments is None:
        instruments = trade_history.Instrument.unique()
    
    for instrument in trade_history.Instrument.unique():
        if str(instrument) != 'nan' and str(instrument) in instruments:
            trade_summary       = trade_history[trade_history.Instrument == instrument]
            
            start_time          = pd.to_datetime(trade_summary.Date.values)[0]
            end_time            = pd.to_datetime(trade_summary.Date.values)[-1]
            formatted_instrument = instrument[:3] + '_' + instrument[-3:]
            
            get_data            = GetData(broker_config)
            aug_start           = start_time - 200*timedelta(seconds = granularity_to_seconds(granularity))
            
            data                = get_data.oanda(formatted_instrument, 
                                                 granularity, 
                                                 start_time = aug_start, 
                                                 end_time = end_time)
            
            # Run AutoTrader in backtest mode to get trade_summary
            home_dir                = os.path.dirname(os.path.abspath(__file__))
            historical_data_name    = 'temp_data.csv'
            historical_data_file_path = os.path.join(home_dir, 
                                                     'price_data',
                                                     historical_data_name)
            data.to_csv(historical_data_file_path)
            
            user_options = {'config_file'   : 'hamacd', 
                            'verbosity'     : '0', 
                            'show_help'     : None, 
                            'notify'        : 0, 
                            'backtest'      : True, 
                            'plot'          : False, 
                            'log'           : False, 
                            'analyse'       : False, 
                            'scan'          : None, 
                            'optimise'      : False, 
                            'data_file'     : historical_data_name,
                            'instruments'   : formatted_instrument
                            }
            
            backtest_trades     = AutoTrader.main(user_options)
            os.remove(historical_data_file_path)
            
            cancelled_orders    = trade_summary[trade_summary.Transaction == 'ORDER_CANCEL']
            insufficient_margin = cancelled_orders[cancelled_orders.Details == 'INSUFFICIENT_MARGIN']
            long_insf_m         = insufficient_margin[insufficient_margin['Direction'] == 'Buy']
            short_insf_m        = insufficient_margin[insufficient_margin['Direction'] == 'Sell']
            
            short_insf_times    = [current_candle(dt, granularity) for dt in pd.to_datetime(short_insf_m.Date.values, utc=True)]
            short_insf_data     = data[data.index.isin(short_insf_times)]
            
            long_insf_times     = [current_candle(dt, granularity) for dt in pd.to_datetime(long_insf_m.Date.values, utc=True)]
            long_insf_data      = data[data.index.isin(long_insf_times)]
            
            filled_orders       = trade_summary[trade_summary.Transaction == 'ORDER_FILL']
            
            sl_orders           = trade_summary[trade_summary.Transaction == 'STOP_LOSS_ORDER']
            tp_orders           = trade_summary[trade_summary.Transaction == 'TAKE_PROFIT_ORDER']
            
            entries             = filled_orders[filled_orders.Details == 'MARKET_ORDER']
            exits               = filled_orders[filled_orders.Details != 'MARKET_ORDER']
            
            long_entries        = entries[entries['Direction'] == 'Buy']
            short_entries       = entries[entries['Direction'] == 'Sell']
            
            profitable_short_exits      = exits[(exits['P/L'] > 0) & (exits['Direction'] == 'Buy')]
            unprofitable_short_exits    = exits[(exits['P/L'] < 0) & (exits['Direction'] == 'Buy')]
            profitable_long_exits       = exits[(exits['P/L'] > 0) & (exits['Direction'] == 'Sell')]
            unprofitable_long_exits     = exits[(exits['P/L'] < 0) & (exits['Direction'] == 'Sell')]
            
            
            ''' Live trade history '''
            # Initialise figure
            fig = make_subplots(rows = 2, cols = 1,
                                shared_xaxes = True,
                                vertical_spacing = 0.02,
                                row_heights = [0.5, 0.5]
                                )
            
            # Price chart
            fig.add_trace(go.Candlestick(x = data.index,
                                         open = data['Open'],
                                         high = data['High'],
                                         low  = data['Low'],
                                         close= data['Close'], 
                                         name = 'market data'
                                         ),
                          row = 1,
                          col = 1
                          )
            
            # Long entries
            fig.add_trace(go.Scatter(x = pd.to_datetime(long_entries.Date.values, utc=True), 
                                     y = long_entries.Price.values, 
                                     mode="markers",
                                     marker_symbol = 105,
                                     marker_size = 12,
                                     marker_line_width = 1,
                                     marker_color = 'black',
                                     name = 'Long position entry'),
                          row = 1,
                          col = 1
                         )
            
            # Cancelled long entries
            fig.add_trace(go.Scatter(x = long_insf_data.index, 
                                      y = long_insf_data.Open.values, 
                                      mode="markers",
                                      marker_symbol = 5,
                                      marker_size = 12,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Cancelled long positions (insufficient margin)'),
                          row = 1,
                          col = 1
                          )
            
            # Short entries
            fig.add_trace(go.Scatter(x = pd.to_datetime(short_entries.Date.values, utc=True), 
                                      y = short_entries.Price.values, 
                                      mode="markers",
                                      marker_symbol = 106,
                                      marker_size = 12,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Short position entry'),
                          row = 1,
                          col = 1
                          )
            
            # # Cancelled short entries
            fig.add_trace(go.Scatter(x = short_insf_data.index, 
                                      y = short_insf_data.Open.values, 
                                      mode="markers",
                                      marker_symbol = 6,
                                      marker_size = 12,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Cancelled short positions (insufficient margin)'),
                          row = 1,
                          col = 1
                          )
            
            # Unprofitable long exits
            fig.add_trace(go.Scatter(x = pd.to_datetime(unprofitable_long_exits.Date.values, utc=True), 
                                     y = unprofitable_long_exits.Price.values, 
                                     mode="markers",
                                     marker_symbol = 0,
                                     marker_size = 5,
                                     marker_line_width = 1,
                                     marker_color = 'black',
                                     name = 'Unprofitable long position exit'),
                          row = 1,
                          col = 1
                          )
            
            # Unprofitable short exits
            fig.add_trace(go.Scatter(x = pd.to_datetime(unprofitable_short_exits.Date.values, utc=True), 
                                      y = unprofitable_short_exits.Price.values, 
                                      mode="markers",
                                      marker_symbol = 0,
                                      marker_size = 5,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Unprofitable short position exit'),
                          row = 1,
                          col = 1
                          )
            
            # Profitable long exits
            fig.add_trace(go.Scatter(x = pd.to_datetime(profitable_long_exits.Date.values, utc=True), 
                                      y = profitable_long_exits.Price.values, 
                                      mode="markers",
                                      marker_symbol = 0,
                                      marker_size = 5,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Profitable long position exit'),
                          row = 1,
                          col = 1
                          )
            
            # Profitable short exits
            fig.add_trace(go.Scatter(x = pd.to_datetime(profitable_short_exits.Date.values, utc=True), 
                                     y = profitable_short_exits.Price.values, 
                                     mode="markers",
                                     marker_symbol = 0,
                                     marker_size = 5,
                                     marker_line_width = 1,
                                     marker_color = 'black',
                                     name = 'Profitable short position exit'),
                          row = 1,
                          col = 1
                         )
            
            # Stop loss levels
            fig.add_trace(go.Scatter(x = pd.to_datetime(sl_orders.Date.values, utc=True), 
                                      y = sl_orders.Price.values, 
                                      mode="markers",
                                      marker_symbol = 41,
                                      marker_size = 12,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Stop loss level'),
                          row = 1,
                          col = 1
                          )
            
            # Take profit levels
            fig.add_trace(go.Scatter(x = pd.to_datetime(tp_orders.Date.values, utc=True), 
                                      y = tp_orders.Price.values, 
                                      mode="markers",
                                      marker_symbol = 41,
                                      marker_size = 12,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Take profit level'),
                          row = 1,
                          col = 1
                          )
            
            # Cancelled stop losses 
            fig.add_trace(go.Scatter(x = pd.to_datetime(insufficient_margin.Date.values, utc=True), 
                                     y = insufficient_margin['Stop Loss'].values, 
                                     mode="markers",
                                     marker_symbol = 41,
                                     marker_size = 12,
                                     marker_line_width = 1,
                                     marker_color = 'black',
                                     name = 'Cancelled short stop losses (insufficient margin)'),
                          row = 1,
                          col = 1
                         )
            
            # Cancelled take profits
            fig.add_trace(go.Scatter(x = pd.to_datetime(insufficient_margin.Date.values, utc=True), 
                                      y = insufficient_margin['Take Profit'].values, 
                                      mode="markers",
                                      marker_symbol = 41,
                                      marker_size = 12,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Cancelled take profits (insufficient margin)'),
                          row = 1,
                          col = 1
                          )
            
            # fig.update_layout(hovermode="y unified")
            fig.update_layout(title = "Trade history chart for {} ({} candles)".format(instrument, 
                                                                                        granularity)
                              )
            
            
            ''' Backtest trade history '''
            # Price chart
            fig.add_trace(go.Candlestick(x      = data.index,
                                         open   = data['Open'],
                                         high   = data['High'],
                                         low    = data['Low'],
                                         close  = data['Close'], 
                                         name   = 'market data'
                                         ),
                          row = 2,
                          col = 1
                          )
            
            # Backtesting signals
            profitable_longs        = backtest_trades[(backtest_trades['Profit'] > 0) 
                                                    & (backtest_trades['Size'] > 0)]
            profitable_shorts       = backtest_trades[(backtest_trades['Profit'] > 0) 
                                                    & (backtest_trades['Size'] < 0)]
            unprofitable_longs      = backtest_trades[(backtest_trades['Profit'] < 0) 
                                                    & (backtest_trades['Size'] > 0)]
            unprofitable_shorts     = backtest_trades[(backtest_trades['Profit'] < 0) 
                                                    & (backtest_trades['Size'] < 0)]
            stop_losses             = backtest_trades.Stop_loss
            take_profits            = backtest_trades.Take_profit
            entry_times             = pd.to_datetime(backtest_trades.Entry_time.values, 
                                                     utc=True)
            exit_times              = list(backtest_trades.Exit_time)
            exit_prices             = backtest_trades.Exit_price
            
            
            # Profitable long trades
            fig.add_trace(go.Scatter(x = pd.to_datetime(profitable_longs.Entry_time.values, utc=True), 
                                     y = profitable_longs.Entry.values, 
                                     mode="markers",
                                     marker_symbol = 5,
                                     marker_size = 12,
                                     marker_line_width = 1,
                                     marker_color = 'lightgreen',
                                     name = 'Profitable long position entry',
                                     text = profitable_longs.Order_ID.values),
                          row = 2,
                          col = 1
                          )
            
            # Profitable short trades
            fig.add_trace(go.Scatter(x = pd.to_datetime(profitable_shorts.Entry_time.values, utc=True), 
                                     y = profitable_shorts.Entry.values, 
                                     mode="markers",
                                     marker_symbol = 6,
                                     marker_size = 12,
                                     marker_line_width = 1,
                                     marker_color = 'lightgreen',
                                     name = 'Profitable short position entry',
                                     text = profitable_shorts.Order_ID.values),
                          row = 2,
                          col = 1
                          )
            
            # Unprofitable long trades
            fig.add_trace(go.Scatter(x = pd.to_datetime(unprofitable_longs.Entry_time.values, utc=True), 
                                     y = unprofitable_longs.Entry.values, 
                                     mode="markers",
                                     marker_symbol = 5,
                                     marker_size = 12,
                                     marker_line_width = 1,
                                     marker_color = 'orangered',
                                     name = 'Unprofitable long position entry',
                                     text = unprofitable_longs.Order_ID.values),
                          row = 2,
                          col = 1
                          )
            
            # Unprofitable short trades
            fig.add_trace(go.Scatter(x = pd.to_datetime(unprofitable_shorts.Entry_time.values, utc=True), 
                                     y = unprofitable_shorts.Entry.values, 
                                     mode="markers",
                                     marker_symbol = 6,
                                     marker_size = 12,
                                     marker_line_width = 1,
                                     marker_color = 'orangered',
                                     name = 'Unprofitable short position entry',
                                     text = unprofitable_shorts.Order_ID.values),
                          row = 2,
                          col = 1
                          )
            
            # Stop loss levels
            fig.add_trace(go.Scatter(x = entry_times, 
                                      y = stop_losses.values, 
                                      mode="markers",
                                      marker_symbol = 41,
                                      marker_size = 12,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Stop loss level'),
                          row = 2,
                          col = 1
                          )
            
            # Take profit levels
            fig.add_trace(go.Scatter(x = entry_times, 
                                      y = take_profits.values, 
                                      mode="markers",
                                      marker_symbol = 41,
                                      marker_size = 12,
                                      marker_line_width = 1,
                                      marker_color = 'black',
                                      name = 'Take profit level'),
                          row = 2,
                          col = 1
                          )
            
            # Position exits
            fig.add_trace(go.Scatter(x = exit_times,
                                     y = exit_prices.values, 
                                     mode="markers",
                                     marker_symbol = 0,
                                     marker_size = 5,
                                     marker_line_width = 1,
                                     marker_color = 'black',
                                     name = 'Position exit',
                                     text = backtest_trades.Order_ID),
                          row = 2,
                          col = 1
                          )
            
            # Hide weekends
            fig.update_xaxes(rangeslider_visible=False,
                             rangebreaks=[
                                 dict(bounds=["sat", "mon"]),
                                 ]
                             )
            fig.show()



def plot_candles(source, fig_tools, height, width):
    
    bull_colour             = "#D5E1DD"
    bear_colour             = "#F2583E"
    candle_colours          = [bear_colour, bull_colour]
    colour_map              = factor_cmap('change', candle_colours, ['0', '1'])
    
    candle_tooltips         = [("Date", "@date{%b %d %H:%M:%S}"),
                               ("Open", "@Open{0.0000}"), 
                               ("High", "@High{0.0000}"), 
                               ("Low", "@Low{0.0000}"),
                               ("Close", "@Close{0.0000}")]

    candle_plot = figure(plot_width     = width, 
                         plot_height    = height, 
                         tools          = fig_tools,
                         active_drag    = 'pan',
                         active_scroll  = 'wheel_zoom')

    candle_plot.segment('index', 'High',
                        'index', 'Low', 
                        color   = "black",
                        source  = source)
    candles = candle_plot.vbar('index', 0.7, 'Open', 'Close', 
                               source       = source,
                               line_color   = "black", 
                               fill_color   = colour_map)
    
    candle_hovertool = HoverTool(tooltips   = candle_tooltips, 
                              formatters    = {'@date':'datetime'}, 
                              mode          = 'mouse',
                              renderers     = [candles])
    
    candle_plot.add_tools(candle_hovertool)
    
    return candle_plot


def plot_macd(x_range, macd_data, linked_fig):
    # Initialise figure
    fig = figure(plot_width     = linked_fig.plot_width,
                 plot_height    = 150,
                 title          = None,
                 tools          = linked_fig.tools,
                 active_drag    = linked_fig.tools[0],
                 active_scroll  = linked_fig.tools[1],
                 x_range        = linked_fig.x_range)

    histcolour = np.where(macd_data['histogram'] < 0, 'red', 'lightblue')
    
    # Add glyphs
    fig.line(x_range, macd_data['macd'], line_color = 'blue')
    fig.line(x_range, macd_data['signal'], line_color = 'red')
    fig.quad(top = macd_data['histogram'],
             bottom = 0,
             left = x_range - 0.3,
             right = x_range + 0.3,
             fill_color = histcolour)

    return fig


def plot_trade_history(data, trade_summary, linked_fig):
    ts = trade_summary
    ts['date']   = ts.index 
    ts           = ts.reset_index(drop = True)
    
    # Here:
    ts_xrange   = data[data.date.isin(trade_summary.index)].index
    
    trade_summary = ts.set_index(ts_xrange)
    data['data_index'] = data.index
    exit_summary = pd.merge(data, trade_summary, left_on='date', right_on='Exit_time')
    
    # Backtesting signals
    profitable_longs        = trade_summary[(trade_summary['Profit'] > 0) 
                                            & (trade_summary['Size'] > 0)]
    profitable_shorts       = trade_summary[(trade_summary['Profit'] > 0) 
                                            & (trade_summary['Size'] < 0)]
    unprofitable_longs      = trade_summary[(trade_summary['Profit'] < 0) 
                                            & (trade_summary['Size'] > 0)]
    unprofitable_shorts     = trade_summary[(trade_summary['Profit'] < 0) 
                                            & (trade_summary['Size'] < 0)]
    
    
    # Profitable long trades
    linked_fig.scatter(list(profitable_longs.index),
                       list(profitable_longs.Entry.values),
                       marker = 'triangle',
                       size = 15,
                       fill_color = 'lightgreen',
                       legend_label = 'Profitable long trades')
    
    # Profitable short trades
    linked_fig.scatter(list(profitable_shorts.index),
                       list(profitable_shorts.Entry.values),
                       marker = 'inverted_triangle',
                       size = 15,
                       fill_color = 'lightgreen',
                       legend_label = 'Profitable short trades')
    
    # Unprofitable long trades
    linked_fig.scatter(list(unprofitable_longs.index),
                       list(unprofitable_longs.Entry.values),
                       marker = 'triangle',
                       size = 15,
                       fill_color = 'orangered',
                       legend_label = 'Unprofitable long trades')
    
    # Unprofitable short trades
    linked_fig.scatter(list(unprofitable_shorts.index),
                       list(unprofitable_shorts.Entry.values),
                       marker = 'inverted_triangle',
                       size = 15,
                       fill_color = 'orangered',
                       legend_label = 'Unprofitable short trades')
    
    
    # Stop loss levels
    linked_fig.scatter(list(trade_summary.index),
                       list(trade_summary.Stop_loss.values),
                       marker = 'dash',
                       size = 15,
                       fill_color = 'black',
                       legend_label = 'Stop loss')
    
    # Take profit levels
    linked_fig.scatter(list(trade_summary.index),
                       list(trade_summary.Take_profit.values),
                       marker = 'dash',
                       size = 15,
                       fill_color = 'black',
                       legend_label = 'Take profit')
    
    # Position exits
    linked_fig.scatter(list(exit_summary.data_index),
                       list(exit_summary.Exit_price.values),
                       marker = 'circle',
                       size = 8,
                       fill_color = 'black',
                       legend_label = 'Position exit')



def plot_portfolio_history(data, balance, NAV, linked_fig, fig_tools):
    # Initialise figure
    fig = figure(plot_width     = linked_fig.plot_width,
                  plot_height    = 150,
                  title          = None,
                  tools          = fig_tools,
                  active_drag    = 'pan',
                  active_scroll  = 'wheel_zoom',
                  x_range        = linked_fig.x_range)
    
    # Add glyphs
    fig.line(data.index, 
             NAV, 
             line_color         = 'black',
             legend_label       = 'Net Asset Value')

    return fig


def plot_indicators(x_range, indicators, linked_fig):
    
    plot_type = {'MACD'        : 'below',
                 'EMA'         : 'over',
                 'RSI'         : 'below',
                 'STOCHASTIC'  : 'below',
                 'SMA'         : 'over',
                 'Heikin-Ashi' : 'below'}
    
    # Plot indicators
    indis_over              = 0
    indis_below             = 0
    max_indis_over          = 3
    max_indis_below         = 2
    bottom_figs             = []
    
    for indicator in indicators:
        indi_type = indicators[indicator]['type']
        
        if indi_type in plot_type:
            if plot_type[indi_type] == 'over' and indis_over < max_indis_over:
    
                linked_fig.line(x_range, 
                                indicators[indicator]['data'], 
                                line_width = 1, 
                                legend_label = indicator)
                indis_over     += 1
                
            elif plot_type[indi_type] == 'below' and indis_below < max_indis_below:
                if indi_type == 'MACD':
                    new_fig     = plot_macd(x_range,
                                            indicators[indicator], 
                                            linked_fig)
                    new_fig.title = indicator
                
                elif indi_type == 'Heikin-Ashi':
                    
                    data        = indicators[indicator]['data']
                    data['date'] = data.index 
                    data        = data.reset_index(drop = True)
                    source      = ColumnDataSource(data)
                    source.add((data.Close >= data.Open).values.astype(np.uint8).astype(str),
                               'change')
                    
                    fig_tools   = "pan,wheel_zoom,box_zoom,undo,redo,reset,save"
                    
                    new_fig     = plot_candles(source, fig_tools, 300, 800)
                    new_fig.x_range = linked_fig.x_range
                    new_fig.title = indicator
                    indis_below   += max_indis_below # To block any other new plots below.
                
                else:
                    new_fig = figure(plot_width     = linked_fig.plot_width,
                                     plot_height    = 130,
                                     title          = None,
                                     tools          = linked_fig.tools,
                                     active_drag    = linked_fig.tools[0],
                                     active_scroll  = linked_fig.tools[1],
                                     x_range        = linked_fig.x_range)
                    
                    # Add glyphs
                    new_fig.line(x_range, 
                                 indicators[indicator]['data'],
                                 line_color         = 'black', 
                                 legend_label       = indicator)
                    
                indis_below    += 1
                bottom_figs.append(new_fig)
        else:
            if indis_below < max_indis_below:
                # Unknown plot type - plot generally on new bottom fig
                print("Indicator not recognised in AutoPlot.")
                new_fig = figure(plot_width     = linked_fig.plot_width,
                                 plot_height    = 130,
                                 title          = None,
                                 tools          = linked_fig.tools,
                                 active_drag    = linked_fig.tools[0],
                                 active_scroll  = linked_fig.tools[1],
                                 x_range        = linked_fig.x_range)
                
                # Add glyphs
                new_fig.line(x_range, 
                             indicators[indicator]['data'],
                             line_color         = 'black', 
                             legend_label       = indicator)
                
                indis_below    += 1
                bottom_figs.append(new_fig)
            
    return bottom_figs


def plot_backtest(backtest_dict):
    data            = backtest_dict['data']
    balance         = backtest_dict['balance']
    NAV             = backtest_dict['NAV']
    trade_summary   = backtest_dict['trade_summary']
    indicators      = backtest_dict['indicators']
    pair            = backtest_dict['pair']
    interval        = backtest_dict['interval']
    
    # Preparation ----------------------------------- #
    output_file("candlestick.html",
                title = "AutoTrader Backtest Results")
    data['date']            = data.index 
    data                    = data.reset_index(drop = True)
    source                  = ColumnDataSource(data)
    source.add((data.Close >= data.Open).values.astype(np.uint8).astype(str),
               'change')
    fig_tools               = "pan,wheel_zoom,box_zoom,undo,redo,reset,save"
    
    # Load JavaScript code for auto-scaling 
    with open(os.path.join(os.path.dirname(__file__), 'lib/autoscale.js'),
              encoding = 'utf-8') as _f:
        autoscale_code      = _f.read()
    
    # Plotting ------------------------------------- #
    # Create main candlestick plot
    candle_plot             = plot_candles(source, fig_tools, 400, 800)
    
    # Overlay trades 
    plot_trade_history(data, trade_summary, candle_plot)
    
    # Plot indicators
    bottom_figs = plot_indicators(data.index, indicators, candle_plot)
    
    # Auto-scale y-axis of candlestick chart
    autoscale_args      = dict(y_range  = candle_plot.y_range, 
                               source   = source)
    candle_plot.x_range.js_on_change('end', CustomJS(args = autoscale_args, 
                                                     code = autoscale_code))
    
    # Above plots
    top_fig             = None
    top_fig             = plot_portfolio_history(data, balance, NAV, candle_plot, fig_tools)
    
    # Compile plots for final figure
    plots               = [top_fig, candle_plot] + bottom_figs
    
    titled  = 0
    t       = Title()
    t.text  = "Backtest chart for {}/{} ({} candles)".format(pair[:3], pair[-3:], interval)
    for plot in plots:
        if plot is not None:
            plot.xaxis.major_label_overrides = {
                i: date.strftime('%b %d') for i, date in enumerate(pd.to_datetime(data["date"]))
            }
            plot.xaxis.bounds   = (0, data.index[-1])
            plot.sizing_mode    = 'stretch_width'
            
            if titled == 0:
                plot.title = t
                titled = 1
            
            if plot.legend:
                plot.legend.visible             = True
                plot.legend.location            = 'top_left'
                plot.legend.border_line_width   = 1
                plot.legend.border_line_color   = '#333333'
                plot.legend.padding             = 5
                plot.legend.spacing             = 0
                plot.legend.margin              = 0
                plot.legend.label_text_font_size = '8pt'
                plot.legend.click_policy        = "hide"
            
            plot.min_border_left    = 0
            plot.min_border_top     = 3
            plot.min_border_bottom  = 6
            plot.min_border_right   = 10
            plot.outline_line_color = 'black'

    # Construct final figure
    fig                 = gridplot(plots, 
                                   ncols            = 1, 
                                   toolbar_location = 'right',
                                   toolbar_options  = dict(logo = None), 
                                   merge_tools      = True
                                   )
    fig.sizing_mode     = 'stretch_width'
    show(fig)



def main(user_options):
    '''
        Main function for generating plots
    '''
    # I want this to be where all the plots are actually generated. 
    # Also, the user options will determine what is plotted. Other functions in
    # this script will be used to generate other plotting bits.
    
    # for example:
    #     if user_options['plot_oanda_trade_history'] is true:
    #         call plot_oanda_trade_history to get the relevant information, 
    #         then proceed with plotting.
    
    
    # Path variables
    home_dir        = os.path.dirname(os.path.abspath(__file__))
    price_data_dir  = os.path.join(home_dir, 'price_data')
    global_config_path = os.path.join(home_dir, 'config', 'GLOBAL.yaml')
    
    granularity     = user_options['granularity']
    
    if user_options['live_performance'] is not None:
        
        broker      = user_options['live_performance']
        filepath    = os.path.join(price_data_dir, user_options['filename'])
        
        global_config   = read_yaml(global_config_path)
        broker_config   = environment_manager.get_config('demo', 
                                                         global_config, 
                                                         broker.upper())
        
        if broker == 'oanda':
            instruments = user_options['instruments']
            plot_oanda_trade_history(filepath, 
                                     global_config, 
                                     broker_config, 
                                     granularity,
                                     instruments)
    
    if user_options['backtest'] is not None:
        plot_backtest(user_options['backtest'])
    
    
    # instrument = "EUR_USD"
    # granularity = 'M30'
    # from_date = datetime()
    # to_date = datetime.now()
    
    # data = []
    
    # # indicators = []
    
    # # Use this dict to either overlay indicator on price chart, or on its own
    # indicator_type = {'RSI': 'indicator',
    #                   'EMA': 'overlay',
    #                   'SMA': 'overlay'
    #                   }
    
    
    # overlays    = len()
    # lowers      = len()
    
    # # Generate chart
    # fig = make_subplots(rows = 1+lowers, cols = 1,
    #                     shared_xaxes = True,
    #                     vertical_spacing = 0.02,
    #                     row_heights = [1]
    #                     )
    
    # # Price chart
    # fig.add_trace(go.Candlestick(x = data.index,
    #                              open = data['Open'],
    #                              high = data['High'],
    #                              low  = data['Low'],
    #                              close= data['Close'], 
    #                              name = 'market data'
    #                              ),
    #               row = 1,
    #               col = 1
    #               )
    
    # for indicator in overlays:
    #     fig.add_trace(go.Scatter(x = data.index, 
    #                              y = indicator, 
    #                              line = dict(color='blue', 
    #                                          width=1
    #                                          ), 
    #                              name = 'name'),
    #                   row = 1,
    #                   col = 1
    #                   )
    
    # for number, indicator in enumerate(lowers):
    #     fig.add_trace(go.Scatter(x = data.index, 
    #                              y = indicator, 
    #                              line = dict(color='blue', 
    #                                          width=1
    #                                          ), 
    #                              name = 'name'),
    #                   row = number + 1,
    #                   col = 1
    #                   )
    
    # fig.show()
    
    return













def print_usage():
    """ Print usage options. """
    print("AutoPlot.py")
    print("Utility to plot price charts and indicators.")
    print("--------------------------------------------------------------" \
          + "---------------")
    print("Flag                                 Comment [short flag]")
    print("--------------------------------------------------------------" \
          + "---------------")
    print("Required:") 
    print("  --granularity                      candlestick granularity [-g]")
    print("  --filename                         data filename [-f]")
    print("\nOptional:")
    print("  --help                             show help for usage [-h]")
    print("  --verbosity <int>                  set verbosity (0,1,2) [-v]")
    print("Coming soon:")
    print("  --instrument 'XXX_YYY'             instrument to stream [-i]")

def print_help(option):
    ''' Print usage instructions. '''
    
    if option == 'instrument' or option == 'i':
        print("Help for '--instrument' (-c) option:")
        
        print("\nExample usage:")
        print("./AutoStream.py -c my_config_file")
        
    elif option == 'verbosity' or option == 'v':
        print("Help for '--verbosity' (-v) option:")
        print("-----------------------------------")
        print("The verbosity flag is used to set the level of output.")


short_options = "g:v:f:h:p:i:"
long_options = ['granularity=', 'verbosity=', 'filename=', 'help=', 
                'performance=', 'instruments=']


if __name__ == '__main__':
    options, r = getopt(sys.argv[1:], 
                        short_options, 
                        long_options
                        )
    
    # Defaults
    instrument      = None
    verbosity       = 0
    filename        = None
    show_help       = None
    live_performance = None
    instruments     = None
    backtest        = None
    
    for opt, arg in options:
        if opt in ('-g', '--granularity'):
            granularity = arg
        elif opt in ('-v', '--verbose'):
            verbosity = arg
        elif opt in ('-f', '--filename'):
            filename = arg
        elif opt in ('-h', '--help'):
            show_help = arg
        elif opt in ('-p', 'live_performance'):
            live_performance = arg
        elif opt in ('-i', 'instruments'):
            instruments = arg
        
    
    uo_dict = {'granularity':   granularity,
               'verbosity':     verbosity,
               'filename':      filename,
               'show_help':     show_help,
               'live_performance': live_performance,
               'instruments':   instruments}

    if len(options) == 0:
        print_usage()
        
    elif uo_dict['show_help'] is not None:
        print_help(uo_dict['show_help'])
        
    else:
        main(uo_dict)






# -*- coding: utf-8 -*-
"""

Dedicated data retrieval class.

"""

import pandas as pd
import v20
import yfinance as yf


class GetData():
    
    def __init__(self, broker_config):
        if broker_config['data_source'] == 'OANDA':
            API                     = broker_config["API"]
            ACCESS_TOKEN            = broker_config["ACCESS_TOKEN"]
            port                    = broker_config["PORT"]
            self.api                = v20.Context(hostname=API, 
                                                  token=ACCESS_TOKEN, 
                                                  port=port)
        
        
        self.base_currency      = None
        

    def oanda(self, instrument, granularity, count=None, start_time=None, end_time=None):
        ''' 
        
            Get historical price data of a instrument from Oanda v20 API.
            
            instrument: the instrument to fetch data for (string)
            
            granularity: candlestick granularity (eg. "M15", "H4", "D")
            
            count: number of candles to fetch, maximum 5000
            
            start_time: data start time (datetime object)
            
            end_time: data end time (datetime object)
            
            If a candlestick count is provided, only one of start time or end
            time should be provided. If neither is provided, the N most
            recent candles will be provided.
            
            Examples:
                data = get_data.oanda("EUR_USD", granularity="M15", start_time=fromdt, end_time=todt)
                data = get_data.oanda("EUR_USD", granularity="M15", start_time=fromdt, count=2110)
                data = get_data.oanda("EUR_USD", granularity="M15", end_time=todt, count=2110)
                data = get_data.oanda("EUR_USD", granularity="M15", count=2110)
        
        '''
        # what if I wanted to request 25,000 candles, rather than specifying 
        # a time range? Would need to modify function again...
        #       Ignore this as an edge case for now. 
        # This would basically be the inverse of what I have already done, 
        # instead of stepping forward, I would step backward with partial_to
        # times until my requested count is hit.
        
        if count is not None:
            # either of count, start_time+count, end_time+count (or start_time+end_time+count)
            # if count is provided, count must be less than 5000
            if start_time is None and end_time is None:
                # fetch count=N most recent candles
                response    = self.api.instrument.candles(instrument,
                                             granularity = granularity,
                                             count = count
                                             )
                data        = self.response_to_df(response)
                
            elif start_time is not None and end_time is None:
                # start_time + count
                from_time   = start_time.timestamp()
                response    = self.api.instrument.candles(instrument,
                                             granularity = granularity,
                                             count = count,
                                             fromTime = from_time
                                             )
                data        = self.response_to_df(response)
            
            elif end_time is not None and start_time is None:
                # end_time + count
                to_time     = end_time.timestamp()
                response    = self.api.instrument.candles(instrument,
                                             granularity = granularity,
                                             count = count,
                                             toTime = to_time
                                             )
                data        = self.response_to_df(response)
                
            else:
                # start_time+end_time+count
                print("Warning: ignoring count input since start_time and",
                       "end_time times have been specified.")
                from_time       = start_time.timestamp()
                to_time         = end_time.timestamp()
            
                # try to get data 
                response        = self.api.instrument.candles(instrument,
                                                         granularity = granularity,
                                                         fromTime = from_time,
                                                         toTime = to_time
                                                         )
                
                # If the request is rejected, max candles likely exceeded
                if response.status != 200:
                    data        = self.get_extended_oanda_data(instrument,
                                                         granularity,
                                                         from_time,
                                                         to_time)
                else:
                    data        = self.response_to_df(response)
                
        else:
            # count is None
            # Assume that both start_time and end_time have been specified.
            from_time       = start_time.timestamp()
            to_time         = end_time.timestamp()
            
            # try to get data 
            response        = self.api.instrument.candles(instrument,
                                                     granularity = granularity,
                                                     fromTime = from_time,
                                                     toTime = to_time
                                                     )
            
            # If the request is rejected, max candles likely exceeded
            if response.status != 200:
                data        = self.get_extended_oanda_data(instrument,
                                                     granularity,
                                                     from_time,
                                                     to_time)
            else:
                data        = self.response_to_df(response)

        return data
    
    
    def get_extended_oanda_data(self, instrument, granularity, from_time, to_time):
        ''' Returns historical data between a date range. '''
        # Currently does not accept count input, but in the future...
        
        max_candles     = 5000
        
        my_int          = self.granularity_to_seconds(granularity)
        end_time        = to_time - my_int
        partial_from    = from_time
        response        = self.api.instrument.candles(instrument,
                                                 granularity = granularity,
                                                 fromTime = partial_from,
                                                 count = max_candles
                                                 )
        data            = self.response_to_df(response)
        last_time       = data.index[-1].timestamp()
        
        while last_time < end_time:
            partial_from    = last_time
            response        = self.api.instrument.candles(instrument,
                                                     granularity = granularity,
                                                     fromTime = partial_from,
                                                     count = max_candles
                                                     )
            
            partial_data    = self.response_to_df(response)
            data            = data.append(partial_data)
            last_time       = data.index[-1].timestamp()
            
        return data
    
    
    def oanda_quote_data(self, data, pair, granularity, start_time, end_time):
        '''
            Function to retrieve price conversion data.
        '''
        quote_currency  = pair[-3:]
        
        if self.base_currency is None or quote_currency == self.base_currency:
            quote_data = data
        else:
            conversion_pair = self.base_currency + "_" + quote_currency
            quote_data = self.oanda(instrument  = conversion_pair,
                                    granularity = granularity,
                                    start_time  = start_time,
                                    end_time    = end_time)
        
        return quote_data
    
    
    def response_to_df(self, response):
        ''' Function to convert api response into a pandas dataframe. '''
        
        candles = response.body["candles"]
        times = []
        close_price, high_price, low_price, open_price = [], [], [], []
        
        for candle in candles:
            times.append(candle.time)
            close_price.append(float(candle.mid.c))
            high_price.append(float(candle.mid.h))
            low_price.append(float(candle.mid.l))
            open_price.append(float(candle.mid.o))
        
        dataframe = pd.DataFrame({"Open": open_price, "High": high_price, "Low": low_price, "Close": close_price})
        dataframe.index = pd.to_datetime(times)
        
        return dataframe
    
    def granularity_to_seconds(self, granularity):
        '''Converts the granularity to time in seconds'''
        letter = granularity[0]
        
        if len(granularity) > 1:
            number = float(granularity[1:])
        else:
            number = 1
        
        conversions = {'S': 1,
                       'M': 60,
                       'H': 60*60,
                       'D': 60*60*24
                       }
        
        my_int = conversions[letter] * number
        
        return my_int
    
    
    def yahoo(self, instrument, granularity=None, start_time=None, end_time=None):
        '''
            Get historical price data from yahoo finance. 
            
            Parameters:
                ticker(s): list of tickers to download
                
                start_string: start time
                
                end_string: end_time
                
                granularity: candlestick granularity
                    valid intervals: 1m, 2m, 5m, 15m, 30m, 60m, 
                                     90m, 1h, 1d, 5d, 1wk, 1mo.
                    note that intraday data cannot extend past 60 days.
                    
            pip install yfinance --upgrade --no-cache-dir
        '''
        
        data = yf.download(tickers  = instrument, 
                           start    = start_time, 
                           end      = end_time,
                           interval = granularity)
        
        return data
    
    def yahoo_quote_data(self, data, pair, interval, from_date, to_date):
        ''' 
            Returns nominal price data - quote conversion not supported for 
            Yahoo finance API.
        '''
        return data
    
    